from .sre_conv import SRE_Conv1d, SRE_Conv2d, SRE_Conv3d, SRE_ConvTranspose2d
from .sre_resnet import SRE_ResNet, sre_resnet18, sre_resnet50
from .sre_resnet import SREBasicBlock, SREBottleneck
from .utils import convert_to_SRE_conv
from .transforms import PadRotateWrapper, FixRotate



from __future__ import print_function
import os
from glob import glob
import shutil
import time
from pytz import timezone
from datetime import datetime
import json
import copy

import numpy as np
import torch
from torch.backends import cudnn
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.cuda.amp import GradScaler
from torch.optim.lr_scheduler import LambdaLR, StepLR, MultiStepLR, CosineAnnealingLR
import torchvision.models.resnet as resnet
import medmnist
from medmnist import INFO
from sklearn.metrics import (
    roc_auc_score,
    accuracy_score,
    balanced_accuracy_score,
    confusion_matrix,
    f1_score,
)
import gc

from opt import get_opt
from SRE_Conv.transforms import get_medmnist_transforms, get_medmnist3d_transforms
import SRE_Conv.sre_resnet as sre_resnet
import SRE_Conv.sre_resnet_3d as sre_resnet_3d
from SRE_Conv.sre_conv import SRE_Conv2d
from utils import (
    log_img,
    manual_seed,
    one_hot,
    log_train_val,
    acc_at_topk,
)


# @profile
def train(args, model, device, train_loader, optimizer, scaler, epoch, log_dir):
    if log_dir:
        train_log_dir = os.path.join(log_dir, "vis_train")
        os.makedirs(train_log_dir, exist_ok=True)
    model.train()
    step_losses = []
    outputs = []
    targets = []
    step_grad_norm = []
    st = time.time()

    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        if len(target.shape) == 2:
            target = target.squeeze(1)
        if args.dev:
            print(data.shape, target.shape)
            print(torch.unique(target.detach().cpu()))
        dtype = torch.bfloat16 if args.bf16 else None
        with torch.autocast("cuda", enabled=args.amp, dtype=dtype):
            output = model(data)
            if args.multi_label:
                loss = F.binary_cross_entropy_with_logits(
                    output, target.to(torch.float32)
                )
            else:
                loss = F.cross_entropy(output, target)
            if args.dev:
                print(data.shape, output.shape, data[:, 0, :10, 0], output)
        # autocast should only wrap the forward pass
        scaler.scale(loss).backward()
        scaler.step(optimizer)
        scaler.update()
        optimizer.zero_grad()
        # cast torch bf16 to f32 for prediction output
        if args.bf16:
            output = output.detach().float()
        else:
            output = output.detach()

        # logging
        step_losses.append(loss.item())
        outputs.append(output.cpu().numpy())
        targets.append(target.detach().cpu().numpy())
        if args.log_grad_norm:
            grad_norm = 0
            for p in model.parameters():
                para_norm = p.grad.data.norm(2)
                grad_norm += para_norm.item() ** 2
            step_grad_norm.append(np.sqrt(grad_norm))
        if batch_idx % args.log_interval == 0:
            print(
                f"Train Epoch: {epoch} [{batch_idx}/{len(train_loader)} ({100. * batch_idx / len(train_loader):.0f}%)]\tLoss: {loss.item():.6f}\tAvg. Loss: {np.mean(step_losses):.6f}\tTime: {time.time()-st:.2f}"
            )
            if log_dir and args.log_img and epoch == 1:
                if data.shape[1] == 3:
                    img = data[0, :].detach().cpu().numpy()
                else:
                    img = data[0, 0].detach().cpu().numpy()
                img_label = target[0].detach().cpu().numpy()
                log_img(epoch, batch_idx, img, img_label, train_log_dir)
            if args.dry_run:
                return step_losses, 0, step_grad_norm
    et = time.time()
    outputs = np.concatenate(outputs, axis=0)
    targets = np.concatenate(targets, axis=0).squeeze()
    # Simply throw away the dummy class
    if args.multi_label:
        preds = (outputs > 0.5).astype(int)
        acc_top5 = 0.0
        auc = 100 * roc_auc_score(targets, outputs, multi_class="ovr")
        f1 = 100 * f1_score(targets, preds, average="macro")
        acc = 100 * accuracy_score(targets.flatten(), preds.flatten())
    else:
        preds = np.argmax(outputs, axis=1).squeeze()
        acc_top5 = 100 * acc_at_topk(targets, outputs, 5)
        auc = 100 * roc_auc_score(
            one_hot(targets, num_classes=outputs.shape[1]), outputs, multi_class="ovr"
        )
        f1 = 100 * f1_score(targets, preds, average="macro")
        acc = 100 * accuracy_score(targets, preds)
    print(
        f"Train set average Acc@1: {acc:.2f}%,\t Acc@5: {acc_top5:.2f}%,\tAUC: {auc:.2f}%,\tF1-score: {f1:.2f}%,\tTime: {et-st:.2f}"
    )
    del targets, outputs
    return step_losses, acc, step_grad_norm


def test(args, model, device, test_loader, epoch, verbose=True, confusion_mat=False):
    model.eval()
    test_loss = 0
    outputs = []
    targets = []
    st = time.time()
    with torch.no_grad():
        for batch_idx, (data, target) in enumerate(test_loader):
            data, target = data.to(device), target.to(device)
            if len(target.shape) == 2:
                target = target.squeeze(1)
            dtype = torch.bfloat16 if args.bf16 else None
            with torch.autocast("cuda", enabled=args.amp, dtype=dtype):
                output = model(data.to(torch.float32))
                if args.multi_label:
                    loss = F.binary_cross_entropy_with_logits(
                        output, target.to(torch.float32), reduction="sum"
                    )
                else:
                    loss = F.cross_entropy(output, target, reduction="sum")
            test_loss += loss.item()  # sum up batch loss
            if args.bf16:
                output = output.detach().float()
            else:
                output = output.detach()
            if args.log_img and epoch == 1 and batch_idx % 10 == 0:
                os.makedirs("./tmp/vis_test", exist_ok=True)
                if data.shape[1] == 3:
                    img = data[0, :].detach().cpu().numpy()
                else:
                    img = data[0, 0].detach().cpu().numpy()
                img_label = target[0].detach().cpu().numpy()
                log_img(0, batch_idx, img, img_label, "./tmp/vis_test")
            outputs.append(output.cpu().numpy())
            targets.append(target.detach().cpu().numpy())
    et = time.time()

    test_loss /= len(test_loader.dataset)
    outputs = np.concatenate(outputs, axis=0)
    targets = np.concatenate(targets, axis=0).squeeze()
    # Simply throw away the dummy class
    if args.multi_label:
        preds = (outputs > 0.5).astype(int)
        acc_top5 = 0.0
        auc = 100 * roc_auc_score(targets, outputs, multi_class="ovr")
        f1 = 100 * f1_score(targets, preds, average="macro")
        ba = 0.0
        acc = 100 * accuracy_score(targets.flatten(), preds.flatten())
    else:
        preds = np.argmax(outputs, axis=1).squeeze()
        acc_top5 = 100 * acc_at_topk(targets, outputs, 5)
        auc = 100 * roc_auc_score(
            one_hot(targets, num_classes=outputs.shape[1]), outputs, multi_class="ovr"
        )
        f1 = 100 * f1_score(targets, preds, average="macro")
        ba = 100 * balanced_accuracy_score(targets, preds)
        acc = 100 * accuracy_score(targets, preds)
    correct = np.sum(targets == preds)

    if verbose:
        print(
            f"\nTest set: Average loss: {test_loss:.4f}, Accuracy: {correct}/{len(test_loader.dataset)} ({acc:.2f}%), Acc@5: {acc_top5:.2f}%, AUC: {auc:.2f}%, BA: {ba:.2f}%, F1-score: {f1:.2f}% Time: {et-st:.2f}\n"
        )
        if confusion_mat:
            print(f"Confusion Matrix:\n{confusion_matrix(targets, preds)}\n")
    return test_loss, acc


def eval_rot(model, test_loader, device, args, verbose=False, eval_3d=False):
    overall_acc = []
    print("## Evaluate on rotation:")
    degrees = range(0, 361, 30) if eval_3d else range(0, 361, 10)
    axes = ["x", "y", "z"] if eval_3d else [""]
    for ax in axes:
        for deg in degrees:
            if verbose:
                print(f"## Evaluate on rotation: {deg} and axis: {ax}")
            args.degree = deg
            args.fix_rotate = True
            if eval_3d:
                args.aug_axis = ax
                _, test_transform = get_medmnist3d_transforms(args)
            else:
                _, test_transform = get_medmnist_transforms(args)
            # hack to test rotation
            test_loader.dataset.transform = test_transform
            test_loss, test_acc = test(
                args, model, device, test_loader, epoch=0, verbose=verbose
            )
            overall_acc.append(round(test_acc, 2))
    print(f"## Overall Acc. for all degrees: {overall_acc}")
    print(
        f"## Overall Avg. Acc.: {np.round(np.mean(overall_acc), 2)}({np.round(np.std(overall_acc), 2)})"
    )
    return np.mean(overall_acc)


def eval_flip(model, test_loader, device, args, verbose=False, eval_3d=False):
    overall_acc = []
    flipping = ["", "v", "h", "d"] if eval_3d else ["", "v", "h"]
    print("## Evaluate on flipping:")
    for flip in flipping:
        if verbose:
            print(f"## Evaluate on flip: {flip}")
        if "v" in flip:
            args.vflip = True
        if "h" in flip:
            args.hflip = True
        if "d" in flip:
            args.dflip = True
        if eval_3d:
            _, test_transform = get_medmnist3d_transforms(args)
        else:
            _, test_transform = get_medmnist_transforms(args)
        # hack to test rotation
        test_loader.dataset.transform = test_transform
        test_loss, test_acc = test(
            args, model, device, test_loader, epoch=0, verbose=verbose
        )
        overall_acc.append(round(test_acc, 2))
        # Reset
        args.vflip = False
        args.hflip = False
        args.dflip = False
    print(f"## Overall Acc. for all flipping: {overall_acc}")
    print(
        f"## Overall Avg. Acc.: {np.round(np.mean(overall_acc[1:]), 2)}({np.round(np.std(overall_acc), 2)})"
    )
    return np.mean(overall_acc)


# @profile
def main_worker(args):
    use_cuda = not args.no_cuda and torch.cuda.is_available()

    manual_seed(args.seed)
    if args.save_model or args.log:
        est = timezone("US/Eastern")
        dt = est.localize(datetime.now())
        dt_str = dt.strftime("%Y-%m-%d-%H-%M-%S")
        task_name = args.med_mnist
        log_dir = os.path.join(
            args.base_log_dir,
            f"{task_name}_{dt_str}_{args.model_type}_{args.exp}_train_logs",
        )
        config_dir = os.path.join(log_dir, "config")
    else:
        log_dir = None

    if use_cuda:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.cudnn:
        cudnn.benchmark = True

    train_kwargs = {"batch_size": args.batch_size}
    test_kwargs = {"batch_size": args.batch_size}
    if use_cuda:
        num_cores = len(os.sched_getaffinity(0))
        if args.num_workers > num_cores:
            args.num_workers = num_cores
        print(f"### Use {args.num_workers} cores for training...")
        cuda_kwargs = {
            "num_workers": args.num_workers,
            "pin_memory": True,
            "shuffle": True,
        }
        train_kwargs.update(cuda_kwargs)
        test_kwargs.update(cuda_kwargs)
    if "ric" in args.model_type:
        train_kwargs["drop_last"] = True
        test_kwargs["drop_last"] = True

    ####### DATASET #######
    in_channels = 3
    info = INFO[args.med_mnist]
    in_channels = info["n_channels"]
    n_classes = len(info["label"])
    if "3d" in args.med_mnist:
        transform, test_transform = get_medmnist3d_transforms(args)
        input_shape = (in_channels, 32, 32, 32)
    else:
        transform, test_transform = get_medmnist_transforms(args)
        input_shape = (in_channels, 32, 32)
    DataClass = getattr(medmnist, info["python_class"])
    train_dataset = DataClass(
        split="train", transform=transform, download=True, size=args.medmnist_size
    )
    test_dataset = DataClass(
        split="test", transform=test_transform, download=True, size=args.medmnist_size
    )

    ####### MODEL #######

    if "3d" in args.med_mnist:
        if args.sre_conv_size_list != None:
            sre_conv_size = args.sre_conv_size_list
        else:
            sre_conv_size = args.sre_conv_size
        model = getattr(sre_resnet_3d, args.model_type)(
            num_classes=n_classes,
            in_channels=in_channels,
            sre_conv_size=sre_conv_size,
            sre_k=args.sre_conv_k,
        )
    elif "sre" in args.model_type:
        if args.sre_conv_size_list != None:
            sre_conv_size = args.sre_conv_size_list
        else:
            sre_conv_size = args.sre_conv_size
        model = getattr(sre_resnet, args.model_type)(
            num_classes=n_classes,
            inplanes=args.res_inplanes,
            sre_conv_size=sre_conv_size,
            large_conv=args.large_conv,
            sre_k=args.sre_conv_k,
            in_channels=in_channels,
        )
        # remove the first downsampling to ensure last stage input size
        if args.res_keep_conv1:
            model.conv1 = SRE_Conv2d(
                in_channels,
                args.res_inplanes,
                kernel_size=5,
                stride=1,
                padding=2,
                bias=False,
            )
        model.maxpool = nn.Identity()
    else:
        model = getattr(resnet, args.model_type)(num_classes=n_classes)
        if in_channels != 3:
            model.conv1 = nn.Conv2d(
                in_channels, 64, kernel_size=7, stride=2, padding=3, bias=False
            )

    if args.maxpool:
        model.avgpool = nn.AdaptiveMaxPool2d((1, 1))

    if args.dev:
        print(model)

    train_loader = torch.utils.data.DataLoader(train_dataset, **train_kwargs)
    test_loader = torch.utils.data.DataLoader(test_dataset, **test_kwargs)
    model = model.to(device)

    #### Scaler
    scaler = GradScaler(enabled=args.scaler)

    param_cnt = sum(
        [param.numel() for param in model.parameters() if param.requires_grad]
    )
    print(f"Correct Total params: {param_cnt}")

    param_group = [{"params": model.parameters()}]
    if args.adam:
        optimizer = optim.Adam(param_group, lr=args.lr, weight_decay=args.weight_decay)
    elif args.sgd:
        optimizer = optim.SGD(
            param_group,
            lr=args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay,
        )
    elif args.adamW:
        optimizer = optim.AdamW(param_group, lr=args.lr, weight_decay=args.weight_decay)
    else:
        optimizer = optim.Adadelta(param_group, lr=args.lr)

    torch.autograd.set_detect_anomaly(True)

    cur_ep = 1
    if args.resume:
        args.log = True
        args.save_model = True
        print(f"### resume experiment logged under {args.resume}...")
        log_dir = args.resume
        config_dir = os.path.join(log_dir, "config")
        ckpt_dir = os.path.join(log_dir, "ckpt")
        assert os.path.exists(log_dir)
        assert os.path.exists(config_dir)
        assert os.path.exists(ckpt_dir)
        args.load_model = log_dir
    elif args.log or args.save_model:
        os.makedirs(log_dir, exist_ok=False)
        print(f"### experiment logged under {log_dir}...")
        os.makedirs(config_dir, exist_ok=True)
        arg_dict = vars(args)
        json.dump(arg_dict, open(os.path.join(config_dir, "train_config.json"), "w"))

    if args.load_model:
        assert os.path.exists(args.load_model)
        if os.path.isdir(args.load_model):
            print(f"### load model logged under {args.load_model}...")
            model_ckpt_dir = os.path.join(args.load_model, "ckpt")
            ckpts = sorted(glob(os.path.join(model_ckpt_dir, "*.ckpt")))
            ckpt_path = ckpts[-1]
        elif os.path.isfile(args.load_model) and ".ckpt" in args.load_model:
            print(f"### load model at {args.load_model}...")
            ckpt_path = args.load_model
        ckpt = torch.load(ckpt_path)
        state_dict = ckpt["state_dict"]
        cur_ep = ckpt["cur_ep"]
        print(f"### load model {ckpt_path} at epoch {cur_ep}...")
        target_state_dict = {}
        for k, param in state_dict.items():
            k = k.replace("module.", "")
            target_state_dict[k] = param
        model.load_state_dict(target_state_dict)

    if cur_ep - 2 != -1:
        # init initial_lr
        for group in optimizer.param_groups:
            group.setdefault("initial_lr", args.lr)
    if args.cos:
        scheduler = CosineAnnealingLR(
            optimizer, T_max=args.epochs, eta_min=args.min_lr, last_epoch=cur_ep - 2
        )
    elif args.step:
        scheduler = StepLR(
            optimizer, step_size=1, gamma=args.gamma, last_epoch=cur_ep - 2
        )
    elif args.multi_step:
        scheduler = MultiStepLR(optimizer, [50, 75], gamma=0.1, last_epoch=cur_ep - 2)
    else:
        scheduler = LambdaLR(optimizer, lambda x: x, last_epoch=cur_ep - 2)

    total_train_loss = []
    total_train_acc = []
    total_grad_norm = []
    total_test_loss = []
    total_test_acc = []
    total_eval_rot_acc = []
    total_eval_flip_acc = []
    best_acc = -1
    for epoch in range(cur_ep, args.epochs + 1):

        if not args.eval_only:
            step_losses, train_acc, grad_norm = train(
                args,
                model,
                device,
                train_loader,
                optimizer,
                scaler,
                epoch,
                log_dir,
            )
            test_loss, test_acc = test(
                args, model, device, test_loader, epoch=epoch, confusion_mat=False
            )
            steps = len(step_losses)
            total_train_loss += step_losses
            total_train_acc += [train_acc for _ in range(steps)]
            if args.log_grad_norm:
                total_grad_norm += grad_norm
            total_test_loss += [test_loss for _ in range(steps)]
            total_test_acc += [test_acc for _ in range(steps)]
            scheduler.step()
            if epoch == 1 and args.dev:
                print(torch.cuda.memory_summary(device=device, abbreviated=False))

            if args.log:
                total_grad_norm = total_grad_norm if args.log_grad_norm else None
                log_train_val(
                    total_train_loss,
                    total_test_loss,
                    total_train_acc,
                    total_test_acc,
                    total_grad_norm,
                    log_dir,
                )

            if args.save_model:
                ckpt_dir = os.path.join(log_dir, "ckpt")
                os.makedirs(ckpt_dir, exist_ok=True)
                if args.save_rec and ((epoch + 1) % args.save_interval) == 0:
                    ckpt = {
                        "state_dict": model.state_dict(),
                        "cur_ep": epoch + 1,
                    }
                    torch.save(
                        ckpt,
                        os.path.join(ckpt_dir, f"{args.model_type}_ep{epoch:0>4d}.ckpt"),
                    )
                elif args.save_best:
                    ckpt_dest = os.path.join(ckpt_dir, f"{args.model_type}_last.ckpt")
                    ckpt = {
                        "state_dict": model.state_dict(),
                        "cur_ep": epoch + 1,
                    }
                    torch.save(ckpt, ckpt_dest)
                    if test_acc > best_acc:
                        best_dist = os.path.join(ckpt_dir, f"{args.model_type}_best.ckpt")
                        print(f"### Update best weight with test auc: {test_acc:.4f}")
                        shutil.copy(ckpt_dest, best_dist)
                        best_acc = test_acc
                else:
                    ckpt_dest = os.path.join(ckpt_dir, f"{args.model_type}_last.ckpt")
                    ckpt = {
                        "state_dict": model.state_dict(),
                        "cur_ep": epoch + 1,
                    }
                    torch.save(ckpt, ckpt_dest)

            gc.collect()

        verbose = (epoch == args.epochs) or args.eval_only
        eval_3d = args.med_mnist != None and "3d" in args.med_mnist
        if args.eval_rot and (epoch % args.eval_interval == 0 or epoch == args.epochs or args.eval_only):
            # don't change the original args
            rot_acc = eval_rot(
                model,
                test_loader,
                device,
                copy.deepcopy(args),
                verbose=verbose,
                eval_3d=eval_3d,
            )
            total_eval_rot_acc.append(rot_acc)

        if args.eval_flip and (epoch % args.eval_interval == 0 or epoch == args.epochs or args.eval_only):
            # don't change the original args
            flip_acc = eval_flip(
                model,
                test_loader,
                device,
                copy.deepcopy(args),
                verbose=verbose,
                eval_3d=eval_3d,
            )
            total_eval_flip_acc.append(flip_acc)

        if args.log and not args.eval_only:
            with open(os.path.join(log_dir, "logs.json"), "w") as f:
                serialize = lambda l: [float(x) for x in l]
                logs = {
                    "train_loss": serialize(total_train_loss),
                    "test_loss": serialize(total_test_loss),
                    "train_acc": serialize(total_train_acc),
                    "test_acc": serialize(total_test_acc),
                    "eval_rot_acc": serialize(total_eval_rot_acc),
                    "eval_flip_acc": serialize(total_eval_flip_acc),
                }
                json.dump(logs, f)
        
        if args.eval_only:
            break


if __name__ == "__main__":
    args = get_opt()
    main_worker(args)



import os
from PIL import Image
import numpy as np
import torch
import random
import matplotlib.pyplot as plt
from pytorch_grad_cam import GradCAM
from pytorch_grad_cam.utils.model_targets import ClassifierOutputTarget
from pytorch_grad_cam.utils.image import show_cam_on_image
from scipy.ndimage import rotate
import imageio
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from scipy.interpolate import interp1d


def one_hot(x, num_classes):
    x = x.squeeze()
    out = np.zeros([x.shape[0], num_classes]).astype(int)
    out[np.arange(x.shape[0]), x] = 1
    return out


def manual_seed(seed=0):
    np.random.seed(seed)
    torch.manual_seed(seed)
    random.seed(seed)

    torch.backends.cudnn.deterministic = True


# @profile
def log_train_val(
    train_loss,
    test_loss=None,
    train_acc=None,
    test_acc=None,
    grad_norm=None,
    log_dir="./log",
):
    plt.figure(figsize=(12, 4))
    plt.plot(np.arange(len(train_loss)), train_loss, label="train_loss")
    if test_loss is not None:
        plt.plot(np.arange(len(test_loss)), test_loss, label="test_loss")
    dest = os.path.join(log_dir, "loss.png")
    plt.legend()
    plt.savefig(dest, dpi=200)
    plt.close("all")

    if train_acc is not None:
        plt.figure(figsize=(12, 4))
        plt.plot(np.arange(len(train_acc)), train_acc, label="train_acc")
        if test_acc is not None:
            plt.plot(np.arange(len(test_acc)), test_acc, label="test_acc")
        dest = os.path.join(log_dir, "acc.png")
        plt.legend()
        plt.savefig(dest, dpi=200)
        plt.close("all")

    if grad_norm is not None:
        plt.figure(figsize=(12, 4))
        plt.plot(np.arange(len(grad_norm)), grad_norm, label="grad_norm")
        dest = os.path.join(log_dir, "grad_norm.png")
        plt.legend()
        plt.savefig(dest, dpi=200)
        plt.close("all")

        plt.figure()
        plt.hist(grad_norm, 10, label="grad_norm_hist")
        percentile_90 = np.percentile(grad_norm, 90)
        dest = os.path.join(log_dir, f"grad_norm_hist_{percentile_90:.4f}.png")
        plt.legend()
        plt.savefig(dest, dpi=200)
        plt.close("all")


def log_img(ep, batch_idx, img, label=None, log_dir="./logs"):
    dest = os.path.join(log_dir, f"ep_{ep}_b_{batch_idx}.png")
    if label is not None:
        dest = dest.replace(".png", f"_label_{label}.png")
    img = (img - img.min()) / (img.max() - img.min())
    if len(img.shape) == 3:
        img = img.transpose((1, 2, 0))
    plt.imsave(dest, img)


def vis_sre_kernel(
    sre_conv, vis_step=1, mode="2D", vmin=None, vmax=None, in_start=0, out_start=0
):
    weight = sre_conv.weight.detach().cpu().numpy()
    kernel = sre_conv._make_weight_matrix(sre_conv.weight).detach().cpu().numpy()
    Cout, Cin, k = weight.shape
    kernel_size = kernel.shape[-1]
    if isinstance(vis_step, int):
        in_vis_step = vis_step
        out_vis_step = vis_step
    else:
        in_vis_step, out_vis_step = vis_step

    in_channels = np.arange(in_start, Cin, in_vis_step)
    out_channels = np.arange(out_start, Cout, out_vis_step)

    if mode == "2D":
        plt.figure(figsize=(2 * len(out_channels), 2 * len(in_channels)))
        idx = 1
        for cin in in_channels:
            for cout in out_channels:
                ax = plt.subplot(len(in_channels), len(out_channels), idx)
                # plt.title(f'K_{cin}_{cout}')
                plt.imshow(kernel[cout, cin, :, :], cmap="Blues", vmin=vmin, vmax=vmax)
                plt.axis("off")
                # cbar = plt.colorbar()

                # if vmin is None:
                #     vmin = kernel[cout, cin, :, :].min()
                # if vmax is None:
                #     vmax = kernel[cout, cin, :, :].max()
                # Apply scientific notation to colorbar
                # # cbar.set_ticks([vmin, (vmin + vmax)/2, vmax])
                # # cbar.set_ticklabels([vmin, (vmin + vmax)/2, vmax])  # Custom labels
                # formatter = ScalarFormatter(useMathText=True)
                # formatter.set_scientific(True)
                # # formatter.set_powerlimits((vmin, vmax))  # You can adjust these limits
                # cbar.ax.yaxis.set_major_formatter(formatter)

                idx += 1
    else:
        plt.figure(figsize=(2 * len(in_channels), 2 * len(out_channels)))
        x = np.arange(k)
        idx = 1
        for cin in in_channels:
            for cout in out_channels:
                ax = plt.subplot(len(in_channels), len(out_channels), idx)
                plt.title(f"K_{cin}_{cout}")
                plt.ylim((vmin, vmax))
                ax.plot(x, weight[cout, cin, :])
                idx += 1
    return plt.gcf()


def create_circle_mask(image_shape, radius):
    # Create grid of coordinates
    x, y = np.ogrid[: image_shape[0], : image_shape[1]]

    center = (image_shape[0] // 2, image_shape[1] // 2)
    # Calculate squared distance from each point to the center
    distance_squared = (x - center[0]) ** 2 + (y - center[1]) ** 2

    # Create mask where True denotes points inside the circle
    mask = distance_squared <= radius**2

    return mask.astype(np.uint8)


def vis_cam(
    nrow,
    ncol,
    model,
    target_layers,
    input_list,
    test_dataset=None,
    img_idx=None,
    save_path=None,
    vis_output=False,
    vis_input=False,
    angle_list=None,
    rot_back=False,
    circle_mask=False,
):
    plt.figure(figsize=(2 * ncol, 2 * nrow))
    plt.margins(x=0)
    for i in range(nrow * ncol):
        if input_list is not None:
            input, label = input_list[i]
        else:
            input, label = test_dataset.__getitem__(img_idx[i])
        if vis_input:
            plt.imsave(
                save_path.replace(".png", f"_{i}_input.png"),
                input.numpy().transpose((1, 2, 0)),
            )
        cam = GradCAM(model=model, target_layers=target_layers)
        input_tensor = input[None]
        targets = [ClassifierOutputTarget(label)]

        grayscale_cam = cam(input_tensor=input_tensor, targets=targets)
        img = input.numpy().transpose((1, 2, 0))
        img = (img - np.min(img)) / (np.max(img) - np.min(img))
        visualization = show_cam_on_image(img, grayscale_cam[0, :], use_rgb=True)

        if angle_list is not None and rot_back:
            angle = angle_list[i]
            h, w, _ = visualization.shape
            visualization = rotate(visualization, -angle, reshape=False)
        if circle_mask:
            h, w, _ = visualization.shape
            mask = create_circle_mask((h, w), 0.5 * h)
            mask = mask == 0
            mask = np.repeat(mask[:, :, np.newaxis], 3, axis=2)
            visualization = np.ma.masked_array(visualization, mask)
            visualization[mask] = 255

        plt.subplot(nrow, ncol, i + 1)
        plt.imshow(visualization)
        plt.axis("off")
    plt.subplots_adjust(wspace=0, hspace=0)
    plt.tight_layout()
    if save_path is not None:
        plt.savefig(save_path, dpi=300, bbox_inches="tight")


outputs = []


def hook(module, input, output):
    global outputs
    outputs.clear()
    outputs.append(output)


outputs = []


def hook_split(module, input, output):
    global outputs
    outputs.clear()
    final_output, temp_output = module._costume_forward_dw(input[0], return_all=True)
    outputs.append(final_output)
    outputs.append(temp_output)


def vis_feat(
    nrow,
    ncol,
    model,
    target_layer,
    input_list,
    feat_idx=1,
    test_dataset=None,
    img_idx=None,
    save_path=None,
    vis_output=False,
    vis_input=False,
    angle_list=None,
    rot_back=False,
    circle_mask=False,
    create_gif=False,
    frame_duration=0.0005,
    gif_input=False,
    rescale=True,
    split_plot=False,
):
    if create_gif:
        img_list = []
    else:
        plt.figure(figsize=(2 * ncol, 2 * nrow))
        plt.margins(x=0)
    for i in range(nrow * ncol):
        if input_list is not None:
            input, label = input_list[i]
        else:
            input, label = test_dataset.__getitem__(img_idx[i])
        if split_plot:
            handle = getattr(model, target_layer)[1].conv1.register_forward_hook(
                hook_split
            )
        elif "layer" in target_layer:
            handle = getattr(model, target_layer)[1].conv1.register_forward_hook(hook)
        else:
            handle = getattr(model, target_layer).register_forward_hook(hook)
        if vis_input:
            plt.imsave(
                save_path.replace(".png", f"_{i}_input.png"),
                input.numpy().transpose((1, 2, 0)),
            )
        output = model(input[None])
        input_img = input[None].detach().numpy().squeeze().transpose((1, 2, 0))
        if split_plot:
            num_rings = outputs[1].shape[1]
            for j in range(num_rings):
                ring_img = outputs[1][:, j].detach().numpy().squeeze()
                ring_img = ring_img.mean(0)
                ring_img = (ring_img - np.min(ring_img)) / (
                    np.max(ring_img) - np.min(ring_img)
                )
                plt.imsave(
                    save_path.replace(".png", f"_{i}_ring_{j}.png"),
                    ring_img,
                    cmap="viridis",
                )
        feat = outputs[0].detach().numpy().squeeze()
        if feat_idx == "mean":
            img = feat.mean(0)
        else:
            assert isinstance(feat_idx, int)
            img = feat[feat_idx]
        img = (img - np.min(img)) / (np.max(img) - np.min(img))
        if angle_list is not None and rot_back:
            angle = angle_list[i]
            h, w = img.shape
            input_img = rotate(input_img, -angle, reshape=False)
            img = rotate(img, -angle, reshape=False)
        if circle_mask:
            h, w = img.shape
            mask = create_circle_mask(img.shape, 0.5 * h)
            mask = mask == 0
            img = np.ma.masked_array(img, mask)
            h, w = input_img.shape[:2]
            mask = create_circle_mask((h, w), 0.5 * h)
            mask = mask == 0
            mask = mask.repeat(3).reshape(h, w, 3)
            input_img = np.ma.masked_array(input_img, mask)

        if create_gif:
            print(f"Creating gif frame: {i+1}/{nrow*ncol}\r", end="")
            if gif_input:
                img_list.append(input_img)
            else:
                # print(img.shape)
                img_list.append(img)
        else:
            plt.subplot(nrow, ncol, i + 1)
            plt.imshow(img, cmap="viridis")
            plt.axis("off")
            plt.imsave(
                save_path.replace(".png", f"_{i}.png"), img, dpi=300, cmap="viridis"
            )
        # plt.imsave(save_path.replace(".png", f"_{i}.png"), img, dpi=300, cmap='viridis')
        handle.remove()
        if vis_output:
            plt.figure()
            plt.bar(
                np.arange(10), torch.softmax(output[0], 0).detach().numpy().squeeze()
            )
            plt.ylim((0, 1))
            plt.savefig(
                save_path.replace(".png", f"_{i}_output.png"),
                dpi=300,
                bbox_inches="tight",
            )
            plt.close("all")
            plt.cla()
    if create_gif:
        cmap = plt.cm.viridis
        output_path = save_path.replace(".png", ".gif")
        # global rescale
        max_intensity = max([img.max() for img in img_list])
        min_intensity = min([img.min() for img in img_list])
        h, w = img_list[0].shape[:2]
        mask = create_circle_mask((h, w), 0.5 * h)
        mask = mask == 0
        if not gif_input:
            mask = np.repeat(mask[:, :, np.newaxis], 4, axis=2)
        with imageio.get_writer(
            output_path, mode="I", duration=frame_duration
        ) as writer:
            for img in img_list:
                if rescale:
                    img = (img - min_intensity) / (max_intensity - min_intensity)
                if gif_input:
                    image_color = img
                else:
                    image_color = cmap(img)
                image_color = (image_color * 255).astype(np.uint8)
                # fill masked region with white color
                image_color[mask] = 255
                writer.append_data(image_color)
        img_stack = np.stack(img_list, axis=0)  # N, H, W
        rotation_wise_std = np.std(img_stack, axis=0)
        # skip the masked region
        if gif_input:
            rotation_wise_std[mask] = 0
        else:
            rotation_wise_std[mask[:, :, 0]] = 0
        avg_std = np.mean(rotation_wise_std)
        print(f"Avg. rotation-wise signal STD: {avg_std:.4f}\n")
        return rotation_wise_std
    else:
        plt.subplots_adjust(wspace=0, hspace=0)
        plt.tight_layout()
        if save_path is not None:
            plt.savefig(save_path, dpi=300, bbox_inches="tight")
        return None


def acc_at_topk(targets, prob, k=5):
    assert len(targets) == len(prob)
    topk = np.argsort(prob, axis=1)[:, -k:]
    correct = 0
    for i in range(len(targets)):
        if targets[i] in topk[i]:
            correct += 1
    return correct / len(targets)


def force_cudnn_initialization():
    s = 32
    dev = torch.device("cuda")
    torch.nn.functional.conv2d(
        torch.zeros(s, s, s, s, device=dev), torch.zeros(s, s, s, s, device=dev)
    )


def visualize_pca_2d(features, labels, vis=True):
    # Perform PCA
    pca = PCA(n_components=2, random_state=42)
    features = features.astype(np.float64)
    # features = (features - features.mean(axis=0)) / features.std(axis=0)
    features = features / np.linalg.norm(features, axis=1)[:, None]
    pca_result = pca.fit_transform(features)
    if not vis:
        return None, pca_result

    # Create a scatter plot of the PCA results
    plt.figure(figsize=(10, 8))
    classes = np.unique(labels)
    colors = plt.cm.tab10(np.linspace(0, 1, len(classes)))

    for i, class_label in enumerate(classes):
        idx = labels == class_label
        plt.scatter(
            pca_result[idx, 0], pca_result[idx, 1], color=colors[i], label=class_label
        )

    plt.title("PCA Visualization of Data")
    plt.xlabel("Principal Component 1")
    plt.ylabel("Principal Component 2")
    plt.xlim(-1.0, 1.0)
    plt.ylim(-1.0, 1.0)
    plt.grid(True)
    return plt.gcf(), pca_result


def visualize_pca_3d(features, labels, vis=True):
    # Perform PCA
    pca = PCA(n_components=3, random_state=42)
    X_pca = pca.fit_transform(features)
    if not vis:
        return None, X_pca

    # Visualize in 3D space
    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111, projection="3d")

    classes = np.unique(labels)
    colors = plt.cm.tab10(
        np.linspace(0, 1, len(classes))
    )  # Assuming there are 3 classes
    for class_label in classes:
        class_indices = np.where(labels == class_label)[0]
        ax.scatter(
            X_pca[class_indices, 0],
            X_pca[class_indices, 1],
            X_pca[class_indices, 2],
            color=colors[class_label],
            label=f"Class {class_label}",
        )

    ax.set_xlabel("Principal Component 1")
    ax.set_ylabel("Principal Component 2")
    ax.set_zlabel("Principal Component 3")
    ax.set_xlim(-1.0, 1.0)
    ax.set_ylim(-1.0, 1.0)
    ax.set_zlim(-1.0, 1.0)
    plt.title("3D PCA Visualization")
    plt.show()


def visualize_tsne_2d(features, labels, vis=True):
    # Perform t-SNE
    tsne = TSNE(
        n_components=2,
        random_state=42,
        init="pca",
        learning_rate="auto",
        n_iter=1500,
        n_iter_without_progress=300,
        verbose=1,
        perplexity=30,
    )
    # tsne = TSNE(n_components=2, random_state=42, init='random', learning_rate='auto',
    #             n_iter=1500, n_iter_without_progress=300, verbose=1, perplexity=20)
    # features = features.astype(np.float64)
    # features = features / np.linalg.norm(features, axis=1)[:, None]
    tsne_result = tsne.fit_transform(features)
    if not vis:
        return None, tsne_result

    # Create a scatter plot of the t-SNE results
    plt.figure(figsize=(10, 8))
    classes = np.unique(labels)
    classes = np.sort(classes)
    colors = plt.cm.tab10(np.linspace(0, 1, len(classes)))

    for i, class_label in enumerate(classes):
        idx = labels == class_label
        plt.scatter(
            tsne_result[idx, 0],
            tsne_result[idx, 1],
            color=colors[i],
            label=class_label,
            s=5,
        )

    plt.title("t-SNE Visualization of Data")
    plt.xlabel("t-SNE Dimension 1")
    plt.ylabel("t-SNE Dimension 2")
    plt.xlim(-110, 110)
    plt.ylim(-110, 110)
    plt.grid(True)
    return plt.gcf(), tsne_result


def visualize_tsne_3d(features, labels, vis=True):
    # Perform t-SNE
    tsne = TSNE(
        n_components=3,
        random_state=42,
        init="pca",
        learning_rate="auto",
        n_iter=1500,
        n_iter_without_progress=300,
        verbose=1,
        perplexity=30,
    )
    # features = features.astype(np.float64)
    # features = features / np.linalg.norm(features, axis=1)[:, None]
    tsne_result = tsne.fit_transform(features)
    if not vis:
        return None, tsne_result

    # Create a scatter plot of the t-SNE results
    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111, projection="3d")
    classes = np.unique(labels)
    classes = np.sort(classes)
    colors = plt.cm.tab10(np.linspace(0, 1, len(classes)))

    for i, class_label in enumerate(classes):
        idx = labels == class_label
        plt.scatter(
            tsne_result[idx, 0],
            tsne_result[idx, 1],
            tsne_result[idx, 2],
            color=colors[i],
            label=class_label,
            s=5,
        )

    plt.title("3D tSNE Visualization")
    ax.set_xlabel("t-SNE Dimension 1")
    ax.set_ylabel("t-SNE Dimension 2")
    ax.set_zlabel("t-SNE Dimension 3")
    ax.set_xlim(-110, 110)
    ax.set_ylim(-110, 110)
    ax.set_zlim(-110, 110)
    return plt.gcf(), tsne_result


def interpolate_1d_array(array, K, axis=1, interp_type="linear"):
    """
    Interpolates a 1D array of size N into K evenly spaced points.

    Parameters:
    array (ndarray): Input ND array of size N.
    K (int): Number of points to interpolate.
    axis (int): Axis along which to interpolate.

    Returns:
    ndarray: Interpolated 1D array of size K.
    """
    N = array.shape[axis]
    x = np.linspace(0, N - 1, N)  # Original indices
    xp = np.linspace(0, N - 1, K)  # New indices

    # Create interpolation function
    interpolator = interp1d(x, array, axis=axis, kind=interp_type)

    # Interpolate the array
    interpolated_array = interpolator(xp)

    return interpolated_array


def create_logits_gif(
    predictions, label, degrees, label2cap, color="tab:blue", dest="./tmp/logits.gif"
):
    n_classes = len(label2cap)
    fig_list = []
    for idx, deg in enumerate(degrees):
        cur_pred = predictions[idx]
        plt.figure(figsize=(10, 6))
        plt.bar(np.arange(n_classes), cur_pred, color=color)
        plt.xticks(np.arange(n_classes), label2cap, rotation=15)
        plt.title(f"Predictions at {deg} degrees")
        plt.ylim((0, 1))
        plt.xlim((-0.5, n_classes - 0.5))
        fig = plt.gcf()
        fig.canvas.draw()
        image_array = np.array(fig.canvas.renderer._renderer)
        plt.close("all")
        fig_list.append(image_array)
    with imageio.get_writer(dest, mode="I", duration=0.002) as writer:
        for img in fig_list:
            writer.append_data(img)
    del fig_list



import torch
import torchio as tio
import torchvision.transforms.functional as F
from torchvision import transforms


class DummyToTensor(object):
    def __init__(self):
        super().__init__()

    def __call__(self, x):
        return torch.tensor(x)


class PadRotateWrapper(object):
    def __init__(self, rotate_trans, padding="zeros", img_size=32):
        super().__init__()
        self.rotate_trans = rotate_trans
        self.padding_mode = padding
        if padding != "zeros":
            self.to_pad = int(img_size // 2)
        else:
            self.to_pad = None
        self.img_size = img_size

    def __call__(self, x):
        # pad the image before rotate
        if self.to_pad != None:
            x = F.pad(
                x,
                (self.to_pad, self.to_pad, self.to_pad, self.to_pad),
                padding_mode=self.padding_mode,
            )
        x = self.rotate_trans(x)
        # crop the image after rotate
        if self.to_pad != None:
            x = F.center_crop(x, [self.img_size, self.img_size])
        return x


class FixRotate(object):
    def __init__(self, degree, expand=False, interpolation="nearest"):
        super().__init__()
        self.degree = degree
        self.expand = expand
        if interpolation == "nearest":
            self.interpolation = F.InterpolationMode.NEAREST
        elif interpolation == "bilinear":
            self.interpolation = F.InterpolationMode.BILINEAR
        elif interpolation == "bicubic":
            self.interpolation = F.InterpolationMode.BICUBIC

    def __call__(self, x):
        return F.rotate(x, self.degree, expand=self.expand,
                        interpolation=self.interpolation)


class PadTransWrapper(object):
    def __init__(self, trans, padding="zeros", img_size=32):
        super().__init__()
        self.trans = trans
        self.padding_mode = padding
        if padding != "zeros":
            self.to_pad = int(img_size // 2)
        else:
            self.to_pad = None
        self.img_size = img_size

    def __call__(self, x):
        # pad the image before rotate
        if self.to_pad != None:
            x = F.pad(
                x,
                (self.to_pad, self.to_pad, self.to_pad, self.to_pad),
                padding_mode=self.padding_mode,
            )
        x = self.trans(x)
        # crop the image after rotate
        if self.to_pad != None:
            x = F.center_crop(x, [self.img_size, self.img_size])
        return x


class FixRotateAxis(object):
    def __init__(self, degree, axis="x", expand=False, interpolation="nearest"):
        super().__init__()
        self.axis = axis
        self.degree = degree
        self.expand = expand
        if interpolation == "nearest":
            self.interpolation = F.InterpolationMode.NEAREST
        elif interpolation == "bilinear":
            self.interpolation = F.InterpolationMode.BILINEAR
        elif interpolation == "bicubic":
            self.interpolation = F.InterpolationMode.BICUBIC

    def __call__(self, x):
        if self.axis == "x":
            x = F.rotate(
                x, self.degree, expand=self.expand, interpolation=self.interpolation
            )
        elif self.axis == "y":
            x = F.rotate(
                x.permute(0, 2, 1, 3),
                self.degree,
                expand=self.expand,
                interpolation=self.interpolation,
            )
            x = x.permute(0, 2, 1, 3)
        elif self.axis == "z":
            x = F.rotate(
                x.permute(0, 3, 1, 2),
                self.degree,
                expand=self.expand,
                interpolation=self.interpolation,
            )
            x = x.permute(0, 2, 3, 1)
        else:
            raise ValueError("axis should be x, y or z")
        return x


class PadTransWrapper3D(PadTransWrapper):
    def __init__(self, trans, padding="zeros", img_size=32):
        super().__init__(trans, padding, img_size)

    def __call__(self, x):
        if self.to_pad != None:
            x = F.pad(
                x,
                (
                    self.to_pad,
                    self.to_pad,
                    self.to_pad,
                    self.to_pad,
                    self.to_pad,
                    self.to_pad,
                ),
                padding_mode=self.padding_mode,
            )
        x = self.trans(x)
        if self.to_pad != None:
            # 3D center crop
            _, H, W, D = x.shape
            Hs = (H - self.img_size) // 2
            Ws = (W - self.img_size) // 2
            Ds = (D - self.img_size) // 2
            x = x[
                :,
                Hs : Hs + self.img_size,
                Ws : Ws + self.img_size,
                Ds : Ds + self.img_size,
            ]
        return x


def get_medmnist_transforms(args):
    target_size = 32 if args.medmnist_size == 28 else args.medmnist_size
    if args.moco_aug:
        transform = [
            transforms.Resize((target_size, target_size)),
            transforms.RandomResizedCrop(target_size, scale=(0.5, 1.0)),
            transforms.RandomApply(
                [transforms.ColorJitter(brightness=0.1, saturation=0.1)],
                p=0.5,  # not strengthened
            ),
            # transforms.RandomGrayscale(p=0.2),
            # transforms.RandomApply([transforms.GaussianBlur((3, 3), [0.1, 2.0])], p=0.2),
            transforms.ToTensor(),
            transforms.Normalize((0.5,), (0.5,)),
        ]
    else:
        transform = [
            transforms.Resize((target_size, target_size)),
            transforms.ToTensor(),
            transforms.Normalize((0.5,), (0.5,)),
        ]

    test_transform = [
        transforms.Resize((target_size, target_size)),
        transforms.ToTensor(),
        transforms.Normalize((0.5,), (0.5,)),
    ]
    if args.fix_rotate:
        test_transform.append(
            PadTransWrapper(
                FixRotate(
                    args.degree, expand=args.expand, interpolation=args.interpolation
                ),
                padding=args.padding,
                img_size=target_size,
            )
        )
    elif args.rotate or args.train_rotate:
        test_transform.append(
            PadTransWrapper(
                transforms.RandomRotation(args.degree, expand=args.expand),
                padding=args.padding,
                img_size=target_size,
            )
        )
    if args.train_rotate and args.fix_rotate:
        transform.append(
            PadTransWrapper(
                FixRotate(
                    args.degree, expand=args.expand, interpolation=args.interpolation
                ),
                padding=args.padding,
                img_size=target_size,
            )
        )
    elif args.train_rotate:
        transform.append(
            PadTransWrapper(
                transforms.RandomRotation(args.degree, expand=args.expand),
                padding=args.padding,
                img_size=target_size,
            )
        )
    if args.translation:
        transform.append(
            PadTransWrapper(
                transforms.RandomAffine(
                    0, (args.translate_ratio, args.translate_ratio)
                ),
                padding=args.padding,
                img_size=target_size,
            )
        )
    if args.test_translation:
        args.test_translate_ratio = (
            args.translate_ratio
            if args.test_translate_ratio is None
            else args.test_translate_ratio
        )
        test_transform.append(
            PadTransWrapper(
                transforms.RandomAffine(
                    0, (args.test_translate_ratio, args.test_translate_ratio)
                ),
                padding=args.padding,
                img_size=target_size,
            )
        )
    if args.vflip:
        test_transform.append(transforms.RandomVerticalFlip(p=1.0))
    if args.hflip:
        test_transform.append(transforms.RandomHorizontalFlip(p=1.0))
    transform = transforms.Compose(transform)
    test_transform = transforms.Compose(test_transform)
    return transform, test_transform


def get_medmnist3d_transforms(args):
    target_size = 32 if args.medmnist_size == 28 else args.medmnist_size
    if args.moco_aug:
        transform = [
            DummyToTensor(),
            tio.transforms.Resize((target_size, target_size, target_size)),
            tio.transforms.RandomAffine(scales=0.2, translation=3, degrees=0),
            tio.transforms.RandomBlur(0.5),
            transforms.Normalize((0.5,), (0.5,)),
        ]
    else:
        transform = [
            DummyToTensor(),
            tio.transforms.Resize((target_size, target_size, target_size)),
            transforms.Normalize((0.5,), (0.5,)),
        ]

    test_transform = [
        DummyToTensor(),
        tio.transforms.Resize((target_size, target_size, target_size)),
        transforms.Normalize((0.5,), (0.5,)),
    ]
    if args.fix_rotate:
        test_transform.append(
            PadTransWrapper3D(
                FixRotateAxis(
                    args.degree,
                    expand=args.expand,
                    axis=args.aug_axis,
                    interpolation=args.interpolation,
                ),
                padding=args.padding,
                img_size=target_size,
            )
        )
    elif args.rotate or args.train_rotate:
        test_transform.append(
            PadTransWrapper3D(
                tio.transforms.RandomAffine(scales=0, degress=args.degree),
                padding=args.padding,
                img_size=target_size,
            )
        )
    if args.train_rotate and args.fix_rotate:
        transform.append(
            PadTransWrapper3D(
                FixRotateAxis(
                    args.degree,
                    expand=args.expand,
                    axis=args.aug_axis,
                    interpolation=args.interpolation,
                ),
                padding=args.padding,
                img_size=target_size,
            )
        )
    elif args.train_rotate:
        transform.append(
            PadTransWrapper3D(
                tio.transforms.RandomAffine(scales=0, degress=args.degree),
                padding=args.padding,
                img_size=target_size,
            )
        )

    if args.vflip:
        test_transform.append(tio.transforms.RandomFlip(axes=1, flip_probability=1.0))
    if args.hflip:
        test_transform.append(tio.transforms.RandomFlip(axes=2, flip_probability=1.0))
    if args.dflip:
        test_transform.append(tio.transforms.RandomFlip(axes=0, flip_probability=1.0))
    transform = transforms.Compose(transform)
    test_transform = transforms.Compose(test_transform)
    return transform, test_transform



import argparse

parser = argparse.ArgumentParser(description="PyTorch MNIST Example")
# Training settings
parser.add_argument(
    "-b",
    "--batch-size",
    type=int,
    default=64,
    metavar="N",
    help="input batch size for training (default: 64)",
)
parser.add_argument(
    "--epochs",
    type=int,
    default=5,
    metavar="N",
    help="number of epochs to train (default: 14)",
)
parser.add_argument(
    "--lr", type=float, default=1.0, metavar="LR", help="learning rate (default: 1.0)"
)
parser.add_argument(
    "--min-lr", type=float, default=1e-6, help="minimal learning rate (default: 1.0)"
)
parser.add_argument(
    "--weight-decay",
    type=float,
    default=1e-3,
    metavar="WD",
    help="weight decay (default: 1e-3)",
)
parser.add_argument(
    "--momentum", type=float, default=0.9, metavar="MT", help="momentum (default: 0.9)"
)
parser.add_argument(
    "--num-workers",
    type=int,
    default=8,
    metavar="NW",
    help="number of workers (default: 8)",
)
parser.add_argument(
    "--gamma",
    type=float,
    default=0.7,
    metavar="M",
    help="Learning rate step gamma (default: 0.7)",
)
parser.add_argument(
    "--step-size", type=float, default=10, help="Step size to update learning rate"
)
parser.add_argument(
    "--no-cuda", action="store_true", default=False, help="disables CUDA training"
)
parser.add_argument(
    "--dry-run", action="store_true", default=False, help="quickly check a single pass"
)
parser.add_argument(
    "--eval-only", action="store_true", default=False, help="quickly evaluate the model"
)
parser.add_argument(
    "--adam", action="store_true", default=False, help="using adam optim"
)
parser.add_argument(
    "--adamW", action="store_true", default=False, help="using adamW optim"
)
parser.add_argument("--sgd", action="store_true", default=False, help="using SGD optim")
parser.add_argument(
    "--cos", action="store_true", default=False, help="using cosine scheduler"
)
parser.add_argument(
    "--step", action="store_true", default=False, help="using step scheduler"
)
parser.add_argument(
    "--step_scheduler-size",
    type=int,
    default=1,
    help="step size for step scheduler (default: 1)",
)
parser.add_argument(
    "--multi-step",
    action="store_true",
    default=False,
    help="using multi-step scheduler",
)
parser.add_argument(
    "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
)
parser.add_argument(
    "--test-batch-size",
    type=int,
    default=1000,
    metavar="N",
    help="input batch size for testing (default: 1000)",
)
parser.add_argument(
    "--model-type", type=str, default="conv", help="which kind of model to use"
)
parser.add_argument(
    "--cudnn",
    action="store_true",
    default=False,
    help="use cudnn benchmark to speed up the training",
)
parser.add_argument("--dev", action="store_true", default=False, help="debug mode")
parser.add_argument("--log-img", action="store_true", default=False, help="log images")
parser.add_argument(
    "--amp", action="store_true", default=False, help="use AMP training"
)
parser.add_argument(
    "--scaler", action="store_true", default=False, help="use gradient scaler training"
)
parser.add_argument(
    "--bf16", action="store_true", default=False, help="use bfloat16 for training"
)

# Log settings
parser.add_argument(
    "--base-log-dir", type=str, default="./logs", help="log output directory"
)
parser.add_argument(
    "--log", action="store_true", default=False, help="log output and train images"
)
parser.add_argument(
    "--log-interval",
    type=int,
    default=200,
    metavar="N",
    help="how many batches to wait before logging training status",
)
parser.add_argument(
    "--save-model",
    action="store_true",
    default=False,
    help="For saving the current Model",
)
parser.add_argument(
    "--log-grad-norm",
    action="store_true",
    default=False,
    help="log model gradient norm",
)
parser.add_argument("--exp", type=str, default="task", help="name of the experiments")
parser.add_argument(
    "--save-rec",
    action="store_true",
    default=False,
    help="For saving the current Model recurrently",
)
parser.add_argument(
    "--save-best",
    action="store_true",
    default=False,
    help="For saving the current best Model",
)
parser.add_argument(
    "--save-interval",
    type=int,
    default=5,
    help="how many epochs to wait before save the model",
)
parser.add_argument("--load-model", type=str, default="", help="load pre-trained model")
parser.add_argument(
    "--resume", type=str, default="", help="resume from previous training"
)
parser.add_argument(
    "--cur-ep", type=int, default=None, help="epoch of previous training"
)


# Augmentation
parser.add_argument(
    "--rotate", action="store_true", default=False, help="rotate the input"
)
parser.add_argument(
    "--train-rotate",
    action="store_true",
    default=False,
    help="rotate the input during training",
)
parser.add_argument(
    "--img-size", type=int, default=128, help="target image size of the input"
)
parser.add_argument(
    "--degree", type=int, default=180, help="degree to rotate during test time"
)
parser.add_argument(
    "--train-degree", type=int, default=30, help="degree to rotate during train time"
)
parser.add_argument(
    "--fix-rotate",
    action="store_true",
    default=False,
    help="rotate the input with fixed angle",
)
parser.add_argument(
    "--moco-aug", action="store_true", default=False, help="using MoCo style transforms"
)
parser.add_argument(
    "--translation", action="store_true", default=False, help="add random translation"
)
parser.add_argument(
    "--test-translation",
    action="store_true",
    default=False,
    help="add random translation",
)
parser.add_argument(
    "--translate-ratio", type=float, default=0.2, help="ratio of random translation"
)
parser.add_argument(
    "--test-translate-ratio",
    type=float,
    default=None,
    help="ratio of random translation at test time",
)
parser.add_argument(
    "--vflip", action="store_true", default=False, help="add vertical flipping"
)
parser.add_argument(
    "--hflip", action="store_true", default=False, help="add horizontal flipping"
)
parser.add_argument(
    "--dflip", action="store_true", default=False, help="add 3d flipping"
)
parser.add_argument(
    "--eval-flip",
    action="store_true",
    default=False,
    help="evaluate the model with flipping",
)
parser.add_argument(
    "--eval-interval", type=int, default=1000, help="interval of specific evaluation"
)
parser.add_argument(
    "--pad-size-h", type=int, default=None, help="height to pad the image"
)
parser.add_argument(
    "--pad-size-w", type=int, default=None, help="weight to pad the image"
)
parser.add_argument(
    "--train-flip-p",
    type=float,
    default=0.5,
    help="probability of apply flipping during training",
)
parser.add_argument(
    "--affine-prob",
    type=float,
    default=0.2,
    help="probability of apply affine transform during training",
)
parser.add_argument(
    "--scale-ratio", type=float, default=0.1, help="ratio to scale the image"
)
parser.add_argument(
    "--shear-ratio", type=float, default=0.1, help="ratio to shear the image"
)
parser.add_argument(
    "--expand",
    action="store_true",
    default=False,
    help="expand the image during rotation",
)
parser.add_argument(
    "--padding",
    type=str,
    default="zeros",
    help="padding method for rotation: zeros | reflect",
)
parser.add_argument(
    "--aug-axis", type=str, default="x", help="axis to apply augmentation: x | y | z"
)
parser.add_argument(
    "--interpolation",
    type=str,
    default="nearest",
    help="interpolation method for rotation: nearest | bilinear | bicubic",
)


# SRE conv settings
parser.add_argument(
    "--sre-shape", type=str, default="o", help="shape of the RI conv kernel"
)
parser.add_argument(
    "--sre-conv-size", type=int, default=3, help="size of the RI convolution kernel"
)
parser.add_argument(
    "--sre-conv-k",
    nargs="+",
    default=None,
    help="# of layers of the RI convolution kernel",
)
parser.add_argument(
    "--res-inplanes",
    type=int,
    default=64,
    help="use different convolution channel size",
)
parser.add_argument(
    "--res-keep-conv1",
    action="store_true",
    default=False,
    help="Not to down sample for resnet Conv1",
)

# Model settings
parser.add_argument(
    "--dropout-p", type=float, default=-1, help="probability of dropout layers"
)
parser.add_argument(
    "--large-conv",
    action="store_true",
    default=False,
    help="Use 7x7 convolution layers",
)
parser.add_argument(
    "--maxpool", action="store_true", default=False, help="use max pooling"
)


# MedMNIST args
parser.add_argument(
    "--sre-conv-size-list",
    nargs="+",
    default=None,
    help="Size of each MNIST ri conv kernel",
)
parser.add_argument(
    "--eval-rot",
    action="store_true",
    default=False,
    help="Evaluate trained model on different rotation degrees",
)
parser.add_argument(
    "--med-mnist", type=str, default="", help="indicate the medMNIST dataset to be used"
)
parser.add_argument(
    "--no-aug", action="store_true", default=False, help="Train w/o augmentation"
)
parser.add_argument(
    "--multi-label",
    action="store_true",
    default=False,
    help="Train with multi-label loss",
)
parser.add_argument(
    "--medmnist-size", type=int, default=28, help="Size of MedMNIST image"
)


def get_opt():
    args = parser.parse_args()
    if args.sre_conv_size_list != None:
        args.sre_conv_size_list = [int(x) for x in args.sre_conv_size_list]
    if isinstance(args.sre_conv_k, list):
        args.sre_conv_k = [int(x) for x in args.sre_conv_k]
    return args



import torch.nn as nn
from .sre_conv import SRE_Conv2d

def convert_to_SRE_conv(model, kernel_shape='o', train_index_mat=False, sre_conv_size=None, sre_groups=None, sre_k=None, force_circular=False):
    '''
    Recursively replace Conv layers to SRE_Conv layers.
    If convolutional layer with stride >= 2 found, add AvgPool2d before it.
    '''
    state = {'counter': 0}

    def _replace_handler(module, state):
        for attr, target in module.named_children():
            if type(target) == nn.Conv2d and target.kernel_size[0] > 1 and target.kernel_size[1] > 1:
                target_padding = target.padding if sre_conv_size is None else target.dilation[0] * (sre_conv_size - 1) // 2
                layer = SRE_Conv2d(
                    target.in_channels,
                    target.out_channels,
                    kernel_size=target.kernel_size if sre_conv_size is None else sre_conv_size,
                    stride=1,
                    padding=target_padding,
                    dilation=target.dilation,
                    groups=target.groups if sre_groups is None else sre_groups,
                    bias=target.bias is not None,
                    sre_k=sre_k,
                    device=target.weight.device,
                    dtype=target.weight.dtype)
                if target.stride[0] > 1 or target.stride[1] > 1:
                    avg_pool = nn.AvgPool2d(target.stride, target.stride)
                    layer = nn.Sequential(avg_pool, layer)
                setattr(module, attr, layer)
                state['counter'] += 1
            _replace_handler(target, state)

    _replace_handler(model, state)